#lang racket/base

(require tests/eli-tester
         racket/match
         redex
         "util/curry.rkt"
         "util/rm.rkt"
         (except-in racket/function const)
         (for-syntax racket/base))

(provide generate-term FRP:restricted evaluate-steps:paper)

#|
Three evaluators:
  ->t - The main evaluator, input is a time step
  ->w1 - phase 1 evaluator for loops, produces a partially evaluated function
  ->w2 - phase 2 evaluator for loops, takes the partially evaluated function and
    produces some output

The main insight of the paper is that feedback loops might be built with
instantaneous cycles thus causing an infinite runtime loop. Feedback built
with decoupled signals get around this issue. By attaching extra type information
to signals the system can detect if a signal is decoupled or not.
|#


#|
;; Things in {} can be optionally specified, if left out agda will try to figure them out
;; implicitly from the context.
;; SF = signal function
;; C = continuous time-signals
;; E = discrete time-signals
prim:
  pure : ∀ {a b} → (a → b) → SF [C a] [C b] cau
  pureE : ∀ {a b } → (a → b) → SF [E a] [E b] cau
  edge : SF [C Bool ] [E Unit] cau
  hold : ∀ {a } → a → SF [E a ] [C a] cau

dprim:                                              
   constant : ∀ {as b} → b → SF as [C b] dec
   never : ∀ {as b } → SF as [E b] dec
   now : ∀ {as } → SF as [E Unit] dec
   pre : ∀ {a } → SF [C a] [C a] dec
|#

(define-language FRP-phase1
  (SF (dprim-phase1 function state))
  (pair SF signals)
  (time integer)
  (function any)
  (state any)
  (signals any))

(define-language FRP:restricted
  (SF (prim prim-function state)
      (dprim dprim-function state)
      (switch SF switch-function)
      (dswitch SF switch-function)
      (loop SF SF)
      (SF op SF))
  (op >>> *** &&&)
  (prim-function prim1 prim2 prim3 prim4)
  (dprim-function dprim1 dprim2 dprim3 dprim4)
  (switch-function SF)
  (state integer))
  
(define-language FRP
  (SF (prim any any)
      (dprim any any)
      (switch SF switch-function)
      (dswitch SF switch-function)
      (loop SF SF)
      (SF op SF))
  (op >>> *** &&&)
  
  (SF-phase1
   (dprim-phase1 any any)
   (loop-phase1 SF-phase1 SF signals)
   (switch-phase1 SF-phase1 function)
   (dswitch-phase1 SF-phase1 function)
   (SF phase1-op SF-phase1))
  (phase1-op >>>-left >>>-right ***-phase1)

  (switch-function any)
  (function any)
  (state any)
  (time integer)
  (signals (signal-part ...))
  (signal-part integer #t #f (just signal-part) nothing unit)
  (value any)
  )

(define-language FRP1
  (SF (prim any any)
      (dprim any)
      (dprim any any)
      (switch SF switch-function)
      (dswitch SF switch-function)
      (loop SF SF)
      (SF op SF)
      (first SF)
      )
  (dSF (dprim any any)
       (SF op dSF)
       (dSF op dSF))
  (E hole
     (time (switch E switch-function) signals)
     (first-E signals E)
     (time (dswitch E switch-function) signals)
     (time (dswitch1 E switch-function) signals)
     (time (loop2 E SF-phase1))
     (time E)
     M ;; is this needed?
     (V op E)
     (E op M)
     
     ;; phase 1
     (loop1 E M)
     (loop3 E time SF signals)
     (E phase1-op SF)
     
     ;; phase 2
     (SF-phase1 phase1-op E)
     (SF phase1-op E)
     (E phase1-op SF-phase1))
  (op >>> *** &&&)
  (switch-function any)
  (M (time SF signals))
  (V (pair SF signals))
  (function any)
  (state any)
  (time integer)
  (signals any)
  (value any)
  (user-functions constant)
  
  ;; phase 1 stuff
  (SF-phase1 (dprim-phase1 any any)
             (loop-phase1 SF-phase1 SF signals)
             (switch-phase1 SF-phase1 function)
             (dswitch-phase1 SF-phase1 function)
             (SF phase1-op SF-phase1))
  (phase1-op >>>-left >>>-right ***-phase1)
  
  (M-phase1 (time SF-phase1))
  (V-phase1 (pair SF-phase1 signals))
  )




;; function : a b -> x
;; (first const) : a b -> (values x b)
;; The result of calling this function with a time and a state is two values,
;; a function and a new state. The new function will accept one argument and always
;; return the first item in the tuple returned by (function t)
(define (make-source function state)
  (term (dprim ,(curry-lambda (t) (compose (first const) (function t))) ,state)))

(define-syntax-rule (debug-it expression)
  (call-with-values (lambda () expression)
                    (lambda args
                      (printf "debug: ~a\n" args)
                      (apply values args))))

(define-metafunction FRP
  maybe-c : -> (prim any any)
  [(maybe-c)
   (prim ,(curry-lambda (a b c) (values (term unit)
                                         (list (term (just unit)))))
         unit)])

(define-metafunction FRP
  constant : any -> (dprim any any)
  [(constant value)
   ,(make-source (curry-lambda (time state)
                           ;; new state
                   (values (term unit)
                           ;; new signals
                           (list (term value))))
                 (term unit))])

(define-metafunction FRP
  delayed : any any -> (dprim any any)
  [(delayed function state_1)
   ,(make-source (curry-lambda (time state)
                               (define new ((term function) state))
                               (values new (list new)))  
                 (term state_1))])

(define-curry (first f a b)
  (values (f a) b))

(define-curry (const a b) a)

(define-curry (edgeAux state signals)
  (let ([first (car signals)])
    (values first
            (if (and first (not state))
                (list (term (just unit)))
                (list (term nothing))))))

(define-metafunction FRP
  edge : -> (prim any any)
  [(edge)
   ;; time is ignored for edge, so we use `const'
   (prim ,(const edgeAux) ,#t)])

;; a function that changes for each invocation
(define (make-stateful function initial) 
  (term (prim ,(curry-lambda (time state signals)
                             (let-values ([(new-state output-signals) (apply function state signals)])
                               (values (term ,new-state)
                                       (term ,(list output-signals)))))
              ,initial)))

(define (make-stateless function)
  (term (prim ,(curry-lambda (time state signals)
                             (values (term unit)
                                     (list (apply function signals))))
              unit)))

(define-metafunction FRP
  second : SF -> SF
  [(second SF)
   ,(let ([swap (term (prim ,(curry-lambda (x1 x2 as)
                               (let ([a (caar as)] [b (cadar as)])
                                 (values (term unit)
                                         (list (list b a)))))
                            unit))])
      (term ((,swap >>> (first SF)) >>> ,swap)))])

(define-metafunction FRP
  unpureC : any any -> (prim any any)
  [(unpureC function state)
   ,(make-stateful (term function) (term state))])

(define-metafunction FRP
  pureC  : any -> (prim any any)
  [(pureC function)
   ,(make-stateless (term function))])

(define-metafunction FRP
  pureE : any -> (prim any any)
  [(pureE function)
   ,(make-stateless (term function))])

(define phase1:reduction
  (reduction-relation
   FRP-phase1
   (--> (time (dprim function state))
        ,(let-values ([(new-state signals)
                       ((term function) (term time) (term state))])
           (term (pair (dprim function ,new-state) ,signals))))
   ))

(define debug? #f)
(define debug-space (make-parameter ""))
(define-syntax-rule (add-debug-space body ...)
  (parameterize ([debug-space (string-append (debug-space) "* ")])
    body ...))
(define-syntax-rule (debug x ...)
  (when debug?
    (display (debug-space))
    (printf x ...)))

(define-syntax-rule (define-reducer name reduction expected-output ...)
  (define-syntax-rule (name what blah (... ...))
     (let ()
       (add-debug-space
         (debug "~a reducing ~a ~a\n" 'name (or blah (... ...) "") (term what))
         (define out (apply-reduction-relation reduction (term what)))
         (debug "~a from ~a got ~a~a\n" 'name (term what) (or blah (... ...) "") out)
         (for ([out1 out])
           (when (not (or (redex-match FRP expected-output out1)
                          ...))
             (error 'name "incorrect result ~a" out1)))
         out))))

(define-reducer reduce paper:reduction (pair SF signals))
(define-reducer reduce-theta1 paper:reduction-theta1 SF-phase1 (pair SF-phase1 signals))
(define-reducer reduce-theta2 paper:reduction-theta2 SF)
                                    
(define paper:reduction-theta1
  (reduction-relation
  FRP
   [--> (time (dprim function state))
        ,(let*-values ([(new-signals->state new-signals) (@ (term function) (term time) (term state))])
           (term (pair (dprim-phase1 function ,new-signals->state) ,new-signals)))]
   [--> (time (SF_sfl >>> SF_sfr))
        (pair (SF_sfl-theta >>>-left SF_sfr*) signals_cs)
        (where ((pair SF_sfl-theta-0 signals_bs-0) ...
                (pair SF_sfl-theta signals_bs)
                (pair SF_sfl-theta-n signals_bs-n) ...)
               ,(reduce-theta1 (time SF_sfl) 'x1))
        (where ((pair SF_sfr*-0 signals_cs-0) ...
                (pair SF_sfr* signals_cs)
                (pair SF_sfr*-n signals_cs-n) ...)
               ,(reduce (time SF_sfr signals_bs) 'x2))]
   [--> (time (SF_sfl >>> SF_sfr))
        (pair (SF_sfl >>>-right SF-phase1_sfr-theta) signals_cs)
        (where ((pair SF_sfr-theta-0 signals_cs-0) ...
                (pair SF-phase1_sfr-theta signals_cs)
                (pair SF_sfr-theta-n+1 signals_cs-n+1) ...)
                ,(reduce-theta1 (time SF_sfr) 'a1))]
   [--> (time (SF_sfl *** SF_sfr))
        (pair (SF_sfl-theta ***-phase1 SF_sfr-theta) ,(append (term signals_cs) (term signals_ds)))
        (where ((pair SF_sfl-theta-0 signals_cs-0) ...
                (pair SF_sfl-theta signals_cs)
                (pair SF_sfl-theta-n signals_cs-n) ...)
               ,(reduce-theta1 (time SF_sfl)))
        (where ((pair SF_sfr-theta-0 signals_ds-0) ...
                (pair SF_sfr-theta- signals_ds)
                (pair SF_sfr-theta-n signals_ds-n) ...)
               ,(reduce-theta1 (time SF_sfr)))]
   [--> (time (loop SF_sfs SF_sff))
        (pair (loop-phase1 SF-phase1_sfs-theta SF_sff* signals_cs) (signal-part_bs ...))
        (side-condition (debug "loop theta1 1\n"))
        (where ((pair SF-phase1_sfs-theta-0 (signal-part_bs-0 ... signal-part_ds-0)) ...
                (pair SF-phase1_sfs-theta (signal-part_bs ... signal-part_ds))
                (pair SF-phase1_sfs-theta-n (signal-part_bs-n ... signal-part_ds-n)) ...)
               ,(reduce-theta1 (time SF_sfs)))
        (side-condition (debug "loop theta1 2\n"))
        (where ((pair SF_sff*-0 signals_cs-0) ...
                (pair SF_sff* signals_cs)
                (pair SF_sff*-n signals_cs-n) ...)
               ,(reduce (time SF_sff (signal-part_ds))))]
   
   [--> (time (switch SF_sf function))
        (pair (switch-phase1 SF_sf-theta function) signals_bs)
        (where ((pair SF_sf-theta-0 (nothing signals_bs-0)) ...
                (pair SF_sf-theta (nothing signals_bs))
                (pair SF_sf-theta-n (nothing signals_bs-n)) ...)
               ,(reduce-theta1 (time SF_sf)))]
   [--> (time (switch SF_sf function))
        (pair SF_sfr-theta signals_bsr)
        (where ((pair (SF_sfs-theta-0 ((just signal-part_e-0) signal-part_bss-0 ...))) ...
                (pair (SF_sfs-theta ((just signal-part_e) signal-part_bss ...)))
                (pair (SF_sfs-theta-n ((just signal-part_e-n) signal-part_bss-n ...))) ...)
               ,(reduce-theta1 (time SF_sf)))
        (where SF_sfr
               ,(@ (term function) (term signal-part_e)))
        (where ((pair SF_sfr-theta-0 signals_bsr-0) ...
                (pair SF_sfr-theta signals_bsr)
                (pair SF_sfr-theta-n signals_bsr-n) ...)
               ,(reduce-theta1 (0 SF_sfr)))]
   
   
   [--> (time (dswitch SF_sf function))
        (pair (dswitch-phase1 SF_sf-theta function) signals_bs)
        (where ((pair SF_sf-theta-0 (nothing signals_bs-0)) ...
                (pair SF_sf-theta (nothing signals_bs))
                (pair SF_sf-theta-n (nothing signals_bs-n)) ...)
               ,(reduce-theta1 (time SF_sf)))]
   [--> (time (dswitch SF_sf function))
        (pair SF_sfr-theta (signal-part_bss ...))
        (where ((pair (SF_sfs-theta-0 ((just signal-part_e-0) signal-part_bss-0 ...))) ...
                (pair (SF_sfs-theta ((just signal-part_e) signal-part_bss ...)))
                (pair (SF_sfs-theta-n ((just signal-part_e-n) signal-part_bss-n ...))) ...)
               ,(reduce-theta1 (time SF_sf)))
        (where SF_sfr
               ,(@ (term function) (term signals_e)))
        (where ((pair SF_sfr-theta-0 signals_bsr-0) ...
                (pair SF_sfr-theta signals_bsr)
                (pair SF_sfr-theta-n signals_bsr-n) ...)
               ,(reduce-theta1 (0 SF_sfr)))]        
   ))

(define paper:reduction-theta2
  (reduction-relation
   FRP
   [--> (time (dprim-phase1 function state_phase1) signals)
        ,(let ([new-state (@ (term state_phase1) (term signals))])
           (term (dprim function ,new-state)))]
   [--> (time (SF_sfl >>>-right SF-phase1_sfr) signals_as)
        (SF_sfl* >>> SF_sfr*)
        (where ((pair SF_sfl*-0 signals_bs-0) ...
                (pair SF_sfl* signals_bs)
                (pair SF_sfl*-n signals_bs-n) ...)
               ,(reduce (time SF_sfl signals_as)))
        (where (SF_sfr*-0 ...
                SF_sfr*
                SF_sfr*-n ...)
               ,(reduce-theta2 (time SF-phase1_sfr signals_bs)))] 
   [--> (time (SF-phase1_sfl >>>-left SF_sfr) signals_as)
        (SF_sfl* >>> SF_sfr)
        (where (SF_sfl*-0 ... SF_sfl* SF_sfl*-n ...)
               ,(reduce-theta2 (time SF_phase1_sfl signals_as)))]
   [--> (time (SF-phase1_sfl ***-phase1 SF-phase1_sfr) (signals_as signals_bs))
        (SF_sfl* *** SF_sfr*)
        (where (SF_sfl*-0 ... SF_sfl* SF_sfr*-n ...)
               ,(reduce-theta2 (time SF-phase1_sfl signals_as)))
        (where (SF_sfr*-0 ... SF_sfr* SF_sfr*-n ...)
               ,(reduce-theta2 (time SF-phase1_sfr signals_bs)))]
   [--> (time (loop-phase1 SF-phase1_sfs SF_sfs* signals_cs) signals_as)
        (loop SF_sfs* SF_sfs*)
        (side-condition (debug "loop-phase1 cs ~a as ~a\n" (term signals_cs) (term signals_as)))
        (where (SF_sfs*-0 ... SF_sfs* SF_sfs*-n ...)
               ,(reduce-theta2 (time SF-phase1_sfs
                                     ,(append (term signals_as) (term signals_cs)))))]
   [--> (time (switch-phase1 SF-phase1_sf function) signals_as)
        (switch SF_sf* function)
        (where (SF_sf*-0 ... SF_sf* SF_sf*-n ...)
               ,(reduce-theta2 (time SF-phase1_sf signals_as)))]
        
   [--> (time (dswitch-phase1 SF-phase1_sf function) signals_as)
        (dswitch SF_sf* function)
        (where (SF_sf*-0 ... SF_sf* SF_sf*-n ...)
               ,(reduce-theta2 (time SF-phase1_sf signals_as)))]
        
   ))

(define paper:reduction
  (reduction-relation
   FRP
   [--> (time (prim function state) signals)
         ,(let*-values ([(new-state new-signal) (@ (term function) (term time) (term state) (term signals))]
                        [(new-signal-function) (term (prim function ,new-state))])
            (term (pair ,new-signal-function ,new-signal)))
         φ-prim]
   
   [--> (time (dprim function state) signals)
        ,(let*-values ([(new-signals->state new-signals) (@ (term function) (term time) (term state))] ;; (f t s)
                       [(new-state) (new-signals->state (term signals))]) ;; z as
           (term (pair (dprim function ,new-state) ,new-signals)))
         φ-dprim]
   
   [--> (time (SF_1 >>> SF_2) signals)
        (pair (SF_1* >>> SF_2*) signals_2)
        (side-condition (debug ">>> 1 signals ~a\n" (term signals)))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_1* signals_1))
               ,(reduce (time SF_1 signals)))
        (side-condition (debug ">>> 2 signals ~a\n" (term signals_1)))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_2* signals_2))
               ,(let ([out (reduce (time SF_2 signals_1))])
                  (debug "out is ~a\n" out)
                  (debug "~a\n" (redex-match FRP ((pair SF_2* signals_2)) out))
                  (debug "~a\n" (redex-match FRP ((pair SF_2* (integer ...))) out))
                  (debug "~a\n" (redex-match FRP ((pair SF_2* any_2)) out))
                  (debug "~a\n"
                         (string?
                         (car
                         (find-binding (match-bindings (car (redex-match FRP ((pair SF_2* any_2)) out)))
                                       'any_2))))
                  out))
        (side-condition (debug ">>> 3 signals ~a\n" (term signals_2)))]
   
   [--> (time (SF_1 &&& SF_2) signals)
        (pair (SF_1* &&& SF_2*) ,(append (term signals_1) (term signals_2)))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_1* signals_1))
               ,(reduce (time SF_1 signals)))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_2* signals_2))
               ,(reduce (time SF_2 signals)))
        ]
   
   [--> (time (SF_1 *** SF_2) signals)
        (pair (SF_1* *** SF_2*) signals_out)
        (where (signal-part_as signal-part_bs rest ...)
               signals)
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_1* signals_cs))
               ,(reduce (time SF_1 (signal-part_as))))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_2* signals_ds))
               ,(reduce (time SF_2 (signal-part_bs))))
        (where signals_out
               ,(append (term signals_cs) (term signals_ds)))]
   
   [--> (time (switch SF function) signals_as)
        (pair (switch SF_out function) (signal-part_out ...))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_out (nothing signal-part_out ...)))
               ,(reduce (time SF signals_as)))]
   
   [--> (time (switch SF_sfs function) signals_as)
        (pair SF_r* signals_bsr)
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_sfs* ((just signal-part_e) signal-part_bss ...)))
               ,(reduce (time SF_sfs signals_as)))
        (where SF_r
               ,((term function) (term signals_e)))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_r* signals_bsr))
               ,(reduce (0 SF_r (signal-part_bss ...))))]
   
   [--> (time (dswitch SF function) signals_as)
        (pair (dswitch SF_out function) (signal-part_out ...))
        (side-condition (debug "dswitch simple ~a\n" (term signals_as)))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_out (nothing signal-part_out ...)))
               ,(reduce (time SF signals_as)))
        (side-condition (debug "dswitch simple SF_out ~a signals ~a\n" (term SF_out) (term (signal-part_out ...))))
        ]

   [--> (time (dswitch SF_sfs function) signals_as)
        (pair SF_r* (signal-part_bss ...))
        (side-condition (debug "dswitch complex ~a\n" (term signals_as)))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_sfs* ((just signal-part_e) signal-part_bss ...)))
               ,(let ()
                 (define out (reduce (time SF_sfs signals_as)))
                 (debug "dswitch complex 1 out ~a\n" out)
                 (debug " SF_sfs*: ~a\n" (redex-match FRP ((pair SF ((just signal-part_e) signal-part_bss ...))) out))
                 out))
        (side-condition (debug "dswitch complex 1 ~a\n" (term SF_sfs*)))
        (where SF_r
               ,((term function) (term (signal-part_e))))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_r* signals_bsr))
               ,(reduce (0 SF_r (signal-part_bss ...))))]       
   
   [--> (time (loop SF_sfs SF_sff) signals_as)
        (pair (loop SF_sfs* SF_sff*) (signal-part_bs))
        (side-condition (debug "loop 0 as ~a\n" (term signals_as)))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF-phase1_sff signals_cs))
               ,(reduce-theta1 (time SF_sff)))
        (side-condition (debug "loop 1 as ~a cs ~a\n" (term signals_as) (term signals_cs)))
        ;; FIXME: use the 0 ... n ... trick
        (where ((pair SF_sfs* (signal-part_bs signal-part_ds ...)))
               ,(let ([ascs (append (term signals_as)
                                    (term signals_cs))])
                  (reduce (time SF_sfs ,ascs))))
        (side-condition (debug "loop 2 ds ~a bs ~a\n" (term (signal-part_ds ...)) (term (signal-part_bs))))

        ;; FIXME: use the 0 ... n ... trick
        (where (SF_sff*)
               ,(reduce-theta2 (time SF-phase1_sff (signal-part_ds ...))))]
        
   ))

(define test-term (term ((1 (constant true) #f) : hole)))

(define (find finder list)
  (let loop ([all list])
    (if (null? all)
        #f
        (let ([found (finder (car all))])
          (if found found (loop (cdr all)))))))

(define (find-binding binds search)
  (find (lambda (bind)
          (and (equal? (bind-name bind) search)
               (bind-exp bind)))
        binds))
                    
;; this can't really be used with the way the reduction relation is set up
;; that is, the code in this module does the big reduction steps as opposed to redex
(define (trace reduction frp signals)
  (traces reduction (term ((1 ,frp ,signals) : hole))))

(define (get-result stuff)
  (let ([matches (redex-match FRP ((pair SF signals) : hole) (car stuff))])
    (when (not matches)
      (error 'get-result "didn't match"))
    (find (lambda (binds)
            (find-binding binds 'signals))
          (map match-bindings matches))))

(define-syntax-rule (define-a-term name stuff)
  (define name (term stuff)))

(define-a-term adder (pureC ,(lambda (x) (+ x 1))))
(define-a-term stringer (pureC ,(lambda (x) (number->string x))))
(define-a-term greater-than-10E (pureE ,(lambda (x) (if (> x 10) (term (just unit)) (term nothing)))))
(define-a-term greater-than-10 (pureC ,(lambda (x) (> x 10))))

(define-a-term integral (unpureC ,(lambda (old signals)
                                    (values (add1 old)
                                            (add1 old)))
                                 ,0))

;; demonstrates signal functions that keep state
(define-a-term fibonacci (unpureC ,(lambda (old signals)
                                     (let ([fib1 (car old)]
                                           [fib2 (cadr old)])
                                       (define new (list (+ fib1 fib2) fib1))
                                       (values new fib2)))
                                  ,(list 1 0)))

(define (pump:paper term signal-maker steps)
  (evaluate-steps:paper term (for/list ([i (in-range 0 steps)]) (signal-maker))))


(define (evaluate:paper frp signals [time 1])
  (reduce (,time ,frp ,signals)))

(define (evaluate-steps:paper term signals [trace? #f])
  (define (doit)
    (for/fold ([output '()]
               [expression term])
      ([time (in-naturals 1)]
       [signal signals])
      (when trace?
        (trace paper:reduction expression signal))
      ;; 1 is the time step, delta-t
      ;; the `time' binding above is absolute time. its not needed
      (define result (match (evaluate:paper expression signal 1)
                       [(list out1) out1]
                       [(list) (error 'evaluate "did not get any results for expression ~a\n" expression)]
                       [rest (error 'evaluate "ambiguous results ~a for ~a\n" rest expression)]))
      (define match1 (redex-match FRP (pair SF signals) result))
      (when (not match1)
        (error 'evaluate "did not get a final result ~a" result))
      (define out (match-bindings (car match1)))
      (values (cons (find-binding out 'signals) output)
              (find-binding out 'SF))))
  (call-with-values doit
                    (lambda (signals expression)
                      (reverse signals))))

(define (get-result:paper stuff)
  (let ([matches (redex-match FRP (pair SF signals) (car stuff))])
    (when (not matches)
      (error 'get-result:paper "didn't match ~a" stuff))
    (find (lambda (binds)
            (find-binding binds 'signals))
          (map match-bindings matches))))

(define-a-term delayed+1 (delayed ,(lambda (state) (+ state 1)) 1))
 
(test
 (get-result:paper (evaluate:paper (term (,adder >>> ,adder)) '(1))) => '(3)
 (get-result:paper (evaluate:paper (term (,adder *** ,adder)) '(1 2))) => '(2 3)

 (get-result:paper (evaluate:paper greater-than-10E '(5))) => (term (nothing))
 (get-result:paper (evaluate:paper greater-than-10E '(12))) => (term ((just unit)))
 (get-result:paper (evaluate:paper greater-than-10 '(5))) => (list #f)
 (get-result:paper (evaluate:paper greater-than-10 '(12))) => (list #t)
 
 (evaluate-steps:paper greater-than-10 '((5) (12))) => '((#f) (#t))
 (evaluate-steps:paper (term (,greater-than-10 >>> (edge))) '((5) (12))) => (term ((nothing)
                                                                                   ((just unit))))
 (evaluate-steps:paper (term (,adder &&& ,adder)) '((1))) => '((2 2))
 (evaluate-steps:paper (term (,greater-than-10 >>> ((edge) &&& (constant 49)))) '((5) (12))) =>
 (term ((nothing 49)
        ((just unit) 49)))
   
 (evaluate-steps:paper (term (switch (,greater-than-10 >>> ((edge) &&& (constant 49)))
                               ,(lambda (e) (term (constant 80)))))
                 '((5) (12) (2))) =>
                              (list (list 49)
                                    (list 80)
                                    (list 80))
                              
 (evaluate-steps:paper (term (dswitch (,greater-than-10 >>> ((edge) &&& (constant 49)))
                               ,(lambda (e) (term (constant 80)))))
                       '((5) (12) (2))) =>
                              (list (list 49)
                                    (list 49)
                                    (list 80))
                              
 (evaluate-steps:paper (term ,integral) '((1) (1) (1))) => '((1) (2) (3))
 (evaluate-steps:paper (term ,fibonacci) '((1) (1) (1) (1) (1) (1)
                                     (1) (1) (1) (1) (1) (1) (1))) =>
                                                                   '((0) (1) (1) (2) (3) (5)
                                                                    (8) (13) (21) (34) (55) (89) (144))
 (pump:paper (term ,fibonacci) (lambda () '(1)) 13) =>
 '((0) (1) (1) (2) (3) (5)
  (8) (13) (21) (34) (55) (89) (144))
 
 
 (evaluate-steps:paper (term (loop (,greater-than-10 *** ,adder)
                             (constant 8)))
                 '((5)))
 => '((#f))
 
 (evaluate-steps:paper (term ,delayed+1) '((1) (1) (1))) => '((2) (3) (4))
 (evaluate-steps:paper (term (loop (,delayed+1 *** ,delayed+1) ,delayed+1))
                       '((1) (1) (1))) => '((2) (3) (4))

)

(provide frp:restricted->frp)
(define (frp:restricted->frp restricted)
  (redex/match FRP:restricted restricted
               [(dprim dprim-function state)
                (case (term dprim-function)
                  [(dprim1) (term (constant state))]
                  [(dprim2) (term (constant state))]
                  [(dprim3) (term (constant state))]
                  [(dprim4) (term (constant state))])]
               [(prim prim-function state)
                (case (term prim-function)
                  [(prim1) (term (pureE ,(curry-lambda (signals) (term state))))]
                  [(prim2) (term (pureE ,(curry-lambda (signals) (term state))))]
                  [(prim3) (term (pureE ,(curry-lambda (signals) (term state))))]
                  [(prim4) (term (pureE ,(curry-lambda (signals) (term state))))])]
               [(SF_1 op SF_2) (term (,(frp:restricted->frp (term SF_1))
                                      op
                                      ,(frp:restricted->frp (term SF_2))))]
               [(loop SF_1 SF_2) (term (loop ,(frp:restricted->frp (term SF_1))
                                             ,(frp:restricted->frp (term SF_2))))]
               [(switch SF switch-function)
                (let ([switch-function* (frp:restricted->frp (term switch-function))])
                  (term (switch ((maybe-c) &&& ,(frp:restricted->frp (term SF)))
                                ,(lambda (e) switch-function*))))]
               [(dswitch SF switch-function)
                (let ([switch-function* (frp:restricted->frp (term switch-function))])
                  (term (dswitch ((maybe-c) &&& ,(frp:restricted->frp (term SF)))
                                 ,(lambda (e) switch-function*))))]
               ))
