#lang scheme

(provide (all-defined-out))

(require
  (prefix-in r: (rename-in redex
                  [apply-reduction-relation                apply-rr₀]
                  [apply-reduction-relation*               apply-rr*₀]
                  [apply-reduction-relation/tag-with-names apply-rr/named₀]))
  (except-in redex
    term define-language define-extended-language
    reduction-relation extend-reduction-relation reduction-relation?
    apply-reduction-relation apply-reduction-relation* traces
    test--> test-->>)
  (for-syntax (only-in redex/reduction-semantics term-match)))
(provide (all-from-out redex))

(require
  (file "util.rkt")
  (for-syntax (file "util.rkt"))
  (for-syntax (except-in syntax/parse id))
  (only-in "unstable/text.rkt" text<?)
  (for-syntax (only-in "unstable/text.rkt" text<? text->identifier)))
(provide test)



; TODO
;  - Observation: Nested 'given's are quadratic!
;  - Hypothesis:  e -ι-> (e e′ ...) -ι-> ((e e′ ...) (e′ e′′ ...) ...) -ι-> ...


;
; Debugging
;

(define *tracing* #f)

(define ((trace-apply-rr f) #:n [n #f] m x)
  (if (not *tracing*)
    (f m x)
    (let ([n (or n (modulo (eq-hash-code m) 1000))]
          [ys (f m x)])
      (printf "  ~s  --~a-->  ~s\n" x n ys)
      ys)))

(define r:apply-rr       (trace-apply-rr r:apply-rr₀))
(define r:apply-rr*      (trace-apply-rr r:apply-rr*₀))
(define r:apply-rr/named (trace-apply-rr r:apply-rr/named₀))

;
; Contracts
;

(define term/c any/c)

;
; Recognize x′ like x_1
;

(define-for-syntax (to-syntax x)
  (datum->syntax #'here x))

; syntax-map : (stx → stx) → (stx list) → (stx list)
(define-for-syntax (syntax-map f stx)
  (to-syntax (map f (syntax->list stx))))

; syntax-concat-map : (stx → (stx list)) → (stx list) → (stx list)
(define-for-syntax (syntax-concat-map f stx)
  (to-syntax (concat-map (compose syntax->list f) (syntax->list stx))))

; add-primes : (stx (id + (list-of id))) → (stx (id + (list-of id)))
(define-for-syntax add-primes
  (syntax-parser
    [(a ...) (syntax-concat-map add-primes #'(a ...))]
    [a       #`(a #,(text->identifier #'a '′))]))

(define-syntax define-language
  (syntax-parser
    [(_ l [nts pat ...] ...)
     (with-syntax ([(nts′ ...) (syntax-map add-primes #'(nts ...))])
       #'(r:define-language l [nts′ pat ...] ...))]))

(define-syntax define-extended-language
  (syntax-parser
    [(_ l′ l [nts pat ...] ...)
     (with-syntax ([(nts′ ...) (syntax-map add-primes #'(nts ...))])
       #'(r:define-extended-language l′ l [nts′ pat ...] ...))]))

; Tests
(define-language k
  [a     number]
  [(b c) any])
(define-extended-language k′ k
  [(b c) number]
  [x     a]
  [(y z) any])
(test-equal (sort (language-nts k) text<?)
            '(a a′ b b′ c c′))
(test-equal (sort (language-nts k′) text<?)
            '(a a′ b b′ c c′ x x′ y y′ z z′))

;
; Models -- replacements for reduction relations
;

; A model is a reduction relation enriched with a section-retraction pair (ι,π):
;
;   rr relation on B x B
;   A -ι-> B -π-> A  =  A -id-> A
;
; The reduction relation operates in the larger space B and the "client" sees
; the reduction system through the lens of the smaller space A.
;
; Since term identity is determined in B, models can separate terms with
; identical A representations. That is, the client might see two separate terms
; in a reduction graph that look the same if the terms are identical in A but
; distinguished in B.
;
(define-struct model (rr ι π n))

(define reduction-relation? model?)

(define (mk-model rr #:ι [ι id] #:π [π id] #:n [n #f])
  (make-model rr ι π n))

(define (set-model-rr rr m) (struct-copy model m [rr rr]))
(define (set-model-ι ι m)   (struct-copy model m [ι ι]))
(define (set-model-π π m)   (struct-copy model m [π π]))
(define (set-model-n n m)   (struct-copy model m [n n]))

(define (add-model-ι ι m) (set-model-ι (compose (model-ι m) ι) m))
(define (add-model-π π m) (set-model-π (compose π (model-π m)) m))
(define (add-model-n n m) (set-model-n n m))

;
; Redex primitives on models instead of reduction relations
;

; reduction-relation : lang rule ... [#:ι ι] [#:π π] -> model
(define-syntax reduction-relation
  (syntax-parser
    [(_ lang x ... #:ι ι y ...)
     #'(add-model-ι ι (reduction-relation lang x ... y ...))]
    [(_ lang x ... #:π π y ...)
     #'(add-model-π π (reduction-relation lang x ... y ...))]
    [(_ lang x ... #:n n y ...)
     #'(add-model-n n (reduction-relation lang x ... y ...))]
    [(_ lang x ...)
     #'(mk-model (r:reduction-relation lang x ...))]))

; extend-reduction-relation : model lang rule ... [#:ι ι] [#:π π] -> model
(define-syntax extend-reduction-relation
  (syntax-parser
    [(_ m-exp lang x ... #:ι ι y ...)
     #'(add-model-ι ι (extend-reduction-relation m-exp lang x ... y ...))]
    [(_ m-exp lang x ... #:π π y ...)
     #'(add-model-π π (extend-reduction-relation m-exp lang x ... y ...))]
    [(_ m-exp lang x ... #:n n y ...)
     #'(add-model-n n (extend-reduction-relation m-exp lang x ... y ...))]
    [(_ m-exp lang x ...)
     #'(let ([m m-exp])
         (mk-model
           (r:extend-reduction-relation (model-rr m) lang x ...)
           #:ι (model-ι m)
           #:π (model-π m)))]))

; traces : model term [options] -> void
(define/kw-dict (traces kws m t)
  (if (not (dict-ref kws '#:pp #f))
    (traces m t #:pp default-pretty-printer)
    (keyword-apply/dict r:traces
      (dict-update kws '#:pp
        (λ (pp)
          (let ([π (model-π m)])
            (if (= (procedure-arity pp) 1)
              (compose pp π)
              (λ (x port width text)
                (pp (π x) port width text))))))
      (model-rr m)
      ((model-ι m) t)
      null)))

; apply-rr : model term -> (list-of term)
(define (apply-rr m t)
  (map (model-π m) (apply-rr/rep m ((model-ι m) t))))

; apply-rr* : model term -> (list-of term)
(define (apply-rr* m t)
  (map (model-π m) (apply-rr*/rep m ((model-ι m) t))))

; apply-rr/named : model term -> (list-of (list (option string) term))
(define (apply-rr/named m t)
  (map (cross id (model-π m)) (apply-rr/named/rep m ((model-ι m) t))))

; apply-rr/rep : model term -> (list-of term)
(define (apply-rr/rep m t)
  (r:apply-rr (model-rr m) t #:n (model-n m)))

; apply-rr*/rep : model term -> (list-of term)
(define (apply-rr*/rep m t)
  (r:apply-rr* (model-rr m) t #:n (model-n m)))

; apply-rr/named/rep : model term -> (list-of (list (option string) term))
(define (apply-rr/named/rep m t)
  (r:apply-rr/named (model-rr m) t))

; nf? : model term -> boolean
(define (nf? m t)
  (empty? (apply-rr m t)))

; test--> : model term term ... -> void
(define-syntax-rule (test--> m e e1 ...)
  (test-equal (apply-rr m e) (list e1 ...)))

; test-->> : model term term ... -> void
(define-syntax-rule (test-->> m e e1 ...)
  (begin
    (test-equal (apply-rr* m e) (list e1 ...))))

;
; Derived forms on models: reduction-relation/* and reduction-relation/given
;

; reduction-relation/* : lang rule [#:ι ι] [#:π π] ... -> model
;
; Dynamic nondeterminism: Like reduction-relation, but expect a list of terms
; instead of a single term on the RHS of the arrow.
;
; For example, here's a toy arithmetic language with a 'drop' operator that
; nondeterministically decreases the argument n by any amount:
;
;   (reduction-relation/* arith
;     [--> (+ n ...) (,(+ (term n) ...))]
;     [--> (drop n) ,(down-from (term n))])
;
; Notice that the deterministic step from (+ n ...) is specified as the
; singleton list (,(+ (term n) ...)).
;
(define-syntax reduction-relation/*
  (syntax-parser
    [(_ lang #:domain d x ...)
     (with-syntax ([ι (gensym 'ι)])
       #'(let* ([m (reduction-relation lang x ...)]
                [n (model-n m)]
                [m (set-model-n (format "[~a]₀" n) m)]
                [apply-rr/rep/* (compose concat apply-rr/rep)])
           (reduction-relation lang #:n n
             #:domain (any d (... ...))
             [--> (ι any_1 any_2 (... ...) any_3 any_4 (... ...))
                  (ι any_3 any_5 (... ...))
                  (where (any_5 (... ...)) ,(apply-rr/rep/* m (term any_3)))]
             #:ι (λ (t) (let ([t′ ((model-ι m) t)])
                          `(ι ,t′ ,@(apply-rr/rep/* m t′))))
             #:π (term-match/single lang
                   [(ι any_1 any_2 (... ...)) ((model-π m) (term any_1))]))))]
    [(_ lang x ...)
     #'(reduction-relation/* lang #:domain any x ...)]))

; reduction-relation/given : lang rule [#:ι ι] [#:π π] ... -> model
;
; Create new reduction relations from old: Like reduction-relation, but allows
; clauses (given rr term pat) that steps 'term' through the reduction relation
; 'rr', binding results with 'pat'. Nondeterminism in 'rr' is preserved into the
; new relation.
;
; This example will add a step counter to an existing relation, preserving any
; nondeterminism already present in the relation:
;
;   (reduction-relation/given lang
;     [--> A     (A 0)]
;     [--> (A n) (B ,(+ 1 (term n)))
;          (given rr A B)])
;
(define-syntax (given stx)
  (raise-syntax-error 'given "used outside of reduction-relation/given"))
(define-syntax reduction-relation/given
  (syntax-parser
    [(_ lang x ...)
     (letrec ([f (syntax-parser #:literals (with given)
                   ; Preserve keywords
                   [(#:domain x y ...) #`(#:domain x #,@(f #'(y ...)))]
                   [(#:arrow  x y ...) #`(#:arrow  x #,@(f #'(y ...)))]
                   [(#:ι      x y ...) #`(#:ι      x #,@(f #'(y ...)))]
                   [(#:π      x y ...) #`(#:π      x #,@(f #'(y ...)))]
                   [(#:n      x y ...) #`(#:n      x #,@(f #'(y ...)))]
                   ; Stop at 'with'
                   [(with x ...)       #`(with x ...)]
                   ; Start 'g' with a singleton list
                   [((arr a a′ x ...) y ...)
                    #`(#,(g #'(arr a (a′) x ...)) #,@(f #'(y ...)))]
                   ; Done
                   [() #'()])]
              [g (syntax-parser #:literals (given)
                   ; (Not sure whether this is left→right or right→left...)
                   [(arr a as′ x ... (given → b b′) y ...)
                    (g #`(arr a ,(concat-map
                                   (∘ concat (term-match lang [b′ (term as′)]))
                                   (apply-rr → (term b)))
                              x ... y ...))]
                   [(arr a as′ x ...)
                    #'(arr a as′ x ...)])])
       #`(reduction-relation/* lang #,@(f #'(x ...))))]))

; Tests
(define-language l
  [(a b c) number])
(define r
  (reduction-relation l #:n 'r
    [--> a ,(* 2 (term a))]
    [--> a ,(+ 1 (* 2 (term a)))]))
(define r*
  (reduction-relation/* l #:n 'r*
    [--> a (,(* 2 (term a)) ,(+ 1 (* 2 (term a))))]))
(define r′
  (reduction-relation/given l #:n 'r′
    [--> a b (given r a b)]))
(define r′′
  (reduction-relation/given l #:n 'r′′
    [--> a b (given r′ a b)]))
(test-equal (sort (apply-rr r   (term 1)) <) (term (2 3)))
(test-equal (sort (apply-rr r*  (term 1)) <) (term (2 3)))
(test-equal (sort (apply-rr r′  (term 1)) <) (term (2 3)))
(test-equal (sort (apply-rr r′′ (term 1)) <) (term (2 3)))

;
; term: mf and meta
;

(define-syntax-rule (mf f)
  (λ xs (term (f ,@xs))))

; (term (meta + 1 2)) --> (r:term ,(apply + (r:term (1 2))))
(define-syntax (meta stx) (raise-syntax-error 'meta "used outside of term"))
(define-syntax (term x)
  (letrec
    ([f (syntax-parser #:literals (meta unquote unquote-splicing)
          [(meta g x ...) #`,(apply g (r:term #,(syntax-map f #'(x ...))))]
          [,e             #`,e]
          [,@e            #`,@e]
          [(t ...)        (syntax-map f #'(t ...))]
          [x              #`x])])
    ; PLT bug 10780: change back to `syntax-parser' and remove arg `x' above
    (syntax-parse x
      [(_ t) #`(r:term #,(f #'t))])))

; Tests
(test-equal (term (meta + 1 2))
            (term 3))
(test-equal (term (f (meta + 1 2)))
            (term (f 3)))
(test-equal (term (f (meta + (meta - 4 3) 2)))
            (term (f 3)))
(test-equal (term (f (meta + (meta - 4 ,(/ 6 2)) 2)))
            (term (f 3)))
(test-equal (term (f (meta + (meta - 4 ,(/ (term 6) 2)) 2)))
            (term (f 3)))
(test-equal (term (f (meta + (meta - 4 ,(/ (term (meta + 3 3)) 2)) 2)))
            (term (f 3)))
(test-equal (term (f (meta + (meta - 4 ,@(snoc '(2) 1)) 2)))
            (term (f 3)))
(test-equal (term (f (meta + (meta - 4 ,@(snoc `(,(term 2)) 1)) 2)))
            (term (f 3)))
(test-equal (term (f (meta + (meta - 4 ,@(snoc `(,(term (meta / 4 2))) 1)) 2)))
            (term (f 3)))

;
; Misc.
;

(define-syntax-rule (match/term lang t x ...)
  (let ([ms ((term-match lang x ...) t)])
    (if (empty? ms)
      (error 'match/term "no matches: ~s" t)
      (first ms))))

(define-syntax-rule (λ/mf lang [pat body] ...)
  (λ xs (match/term lang xs [pat (term body)] ...)))

(define (tests-predicate p xs)
  (for-each (λ (x) (test-predicate p x)) xs))

(define-syntax-rule (tests-mapsto f (x ... y) ...)
  (begin
    (test-equal (f x ...) y)
    ...))

(define-syntax-rule (tests-redex-match l p a ...)
  (begin
    (test-predicate id (redex-match l p (term a)))
    ...))

(define-syntax-rule (tests-redex-no-match l p a ...)
  (begin
    (test-predicate not (redex-match l p (term a)))
    ...))

(define-syntax-rule (tests-metafunction-mapsto f (x ... y) ...)
  (begin
    (test-equal (term (f x ...)) (term y))
    ...))

(define-syntax-rule (tests--> r (a b ...) ...)
  (begin
    (test--> r (term a) (term b) ...)
    ...))

(define-syntax-rule (tests-->> r (a b ...) ...)
  (begin
    (test-->> r (term a) (term b) ...)
    ...))
