#lang scheme

(require redex/reduction-semantics
         "source.rkt"
         "target.rkt"
         ;"target-tests.rkt"
         "compiler.rkt"
         "can-reach.rkt"
         "hashed-reduction-relation.rkt"
         "pi-calculus.rkt")

;; This is the top level file: it tries to verify the claim made by the paper.


(define arr/t (apply-reduction-relation/hashed (lambda (t) (apply-reduction-relation t-> t))))

(define checker/t (lambda (t) (let ([a1 (apply-reduction-relation t-> t)]
                                    [a2 (arr/t t)])
                                (unless (equal? a1 a2)
                                  (error 'checker/t "yikes! didn't get the same answer for a1 and a2"))
                                a1)))

(define bfs-max-terms 10)
(define twiddle-max-depth 3)


;; THE PAPER'S CLAIM:    S  ~  COMPILE(S)

;; this is going to diverge for every term in existence.

(define twiddle-hash (make-hash))

;; the twiddle relation contains four clauses: invariance, safety, progress, and fairness.
(define (twiddle-related? s-term t-term #:max-depth [max-depth #f])
  (printf "entering twiddle-related with ~v and ~v and ~v\n" s-term t-term max-depth)
  (let ([hashed (hash-ref twiddle-hash s-term (lambda () null))])
    (cond [(member t-term hashed) (begin (printf "hash hit!\n") #t)]
          [(and max-depth (<= max-depth 0))
           (begin (printf "twiddle depth exceeded, returning #t\n")
                  #t)]
          [else (and (invariance-holds? s-term t-term)
                     (safety-holds?     s-term t-term #:max-depth max-depth)
                     (progress-holds?   s-term t-term #:max-depth max-depth)
                     (fairness-holds?   s-term t-term #:max-depth max-depth)
                     (hash-set! twiddle-hash s-term (cons t-term hashed)))])))


;; INVARIANCE : can you take steps in the machine to get to something 
;;   with the same denotation as the source?

(define (invariance-holds? prog state)
  (with-handlers ([exn:search-failure? (lambda (exn) (printf  "search failure: couldn't take steps in ~v to reach something with the same denotation as ~v\n"
                                                              state prog)
                                         #f)])
    (can-reach/pred checker/t state (lambda (s2) (denotes-equal? prog s2)) #:max-terms bfs-max-terms)))


;; SAFETY : if there's a machine step, zero or more steps on the source
;;  side will restore twiddle-hood.

(define (safety-holds? s-term t-term #:max-depth [max-depth #f])
  (let* ([next-t-steps (checker/t t-term)]
         [twiddle-equivalent-preds (map (lambda (new-t-term) (lambda (s-term) (twiddle-related? s-term new-t-term #:max-depth (and max-depth (- max-depth 1))))) 
                                        next-t-steps)])
  (apply can-reach/pred (lambda (s) (apply-reduction-relation s-> s)) s-term twiddle-equivalent-preds #:max-terms bfs-max-terms)))



;; PROGRESS: if the source takes a step, then it's possible to take a step on the 
;;  source side and one or more steps on the machine side and get twiddlehood.

;; the big question with the statement of this clause is why it doesn't have to be
;; true for every step possible on the source side.  Are there steps possible in the 
;; source that the target can't take? ... Ah, Yes! because the machine state may have
;; committed to trying something, and therefore can't do the thing that the source does.

(define (progress-holds? s-term t-term #:max-depth [max-depth #f])
  (let ([next-s-terms (apply-reduction-relation s-> s-term)])
    (or (empty? next-s-terms)
        ;; note the *OR*, not an *AND* :
        (for/or ([ns (in-list next-s-terms)])
          (can-reach/pred/1+ checker/t t-term (lambda (t)
                                                (printf "call from the progress-holds\n")
                                                (twiddle-related? ns t #:max-depth (and max-depth (- max-depth 1))))
                             #:max-terms bfs-max-terms)))))



;; FAIRNESS: if the source takes a step, then it's possible to find some machine state that 
;;   reduces to the given machine state where all but the last step of the sequence are twiddle equivalent
;;   to the given source step, and that also reduces via *one or more* steps to a state equivalent
;;   to the result of the source step.

;; In all honesty, it's very unclear to me how this constitutes fairness. It seems to me that fairness should
;; state that if the freshly compiled source term can reduce to a state with some denotation, then the 
;; machine term should also be able to get there.  This doesn't hold for smaller steps, but I don't see
;; that the listed property even implies this.

;; um, besides being essentially impossible to check in a quickcheck-style way.

(define (fairness-holds? s-term t-term #:max-depth [max-depth #f])
  (for/and ([ns (in-list (apply-reduction-relation s-> s-term))])
    (find-t-term-such-that
     (lambda (past-t-term)
       ;; DEAD CODE: this body is never evaluated; it's believed to be correct, though,
       ;; if find-t-term-such-that actually did something useful.
       (and (can-reach checker/t past-t-term t-term)
            (let ([seq (path-from-to/reversed t-> past-t-term t-term)])
              (and (for/and ([t (in-list (rest seq))])
                     (twiddle-related? s-term t))
                   (can-reach/pred/1+ checker/t past-t-term (lambda (t) (twiddle-related? ns t #:max-depth (and max-depth (- max-depth 1)))))))))
     #:hint t-term)))


;; find a machine state that satisfies the predicate 
;; by trying all possible states. Can't possibly fail, 
;; instead would just run forever... since we're looking
;; for counterexamples, we give up right here. 

;; well, we might as well try the hint, anyway.

(define (find-t-term-such-that pred #:hint [hint #f])
  (cond [(and hint (pred hint)) hint]
        [else 
         (printf "giving up, assuming term exists.\n")
         'not-a-real-term]))



;; do the given source term and machine state have the same denotation?
(define (denotes-equal? s-term t-term)
  (equal? (sort-denotation (term (denotes ,s-term)))
          (sort-denotation (term (denotes/t ,t-term)))))

;; sort denotations... turns out to be easiest to do using string-sort!
(define (sort-denotation s)
  (sort s string<? #:key (lambda (x) (format "~e" x))))





;; Does the main property hold? uncomment this to see it spin on terms like ((select (in c1) (out c1)))

#;(redex-check sourcelang prog (twiddle-related? (term prog) (term (compile prog))) #:attempts 1000)



;; SANITY CHECK 

;; do freshly compiled terms satisfy invariance? (yes, of course, no reductions required)
#;(time (redex-check sourcelang prog (invariance-holds? (term prog) (term (compile prog))) #:attempts 100))



;; OBSERVATION: TERMS WITHOUT REDUCTIONS COMPILE INTO TERMS WITH REDUCTIONS. CAN REDEX FIND THEM? Yes.

;; does the given term have no steps?
(define (has-no-steps? relation t) (equal? '() (apply-reduction-relation relation t)))
(define (implies a b) (if a b #t))

#;(time (redex-check sourcelang prog (implies (has-no-steps? s-> (term prog)) (has-no-steps? t-> (term (compile prog)))) #:attempts 100))



;; GETTING A HANDLE ON IT: RULING OUT BADLY BEHAVED TERMS

;; Terms that contain a select with both an 'in' and an 'out' on the same channel will spin. Let's rule them out.

;; does the source term contain a select with a matching in/out pair?
(define-metafunction sourcelang
  select-inout-pair : prog -> variable
  [(select-inout-pair (s_1 ... (select a_1 ... (in chan) a_2 ... (out chan) a_3 ...) s_2 ...)) t]
  [(select-inout-pair (s_1 ... (select a_1 ... (out chan) a_2 ... (in chan) a_3 ...) s_2 ...)) t]
  [(select-inout-pair (s_1 ...)) f])

(test-equal (term (select-inout-pair ((in c1) (out c2) (select (in A) (out B)) (select (out A) (in B) (out A))))) (term f))
(test-equal (term (select-inout-pair ((in c1) (out c2) (select (in A) (out B)) (select (out A) (in B) (in A))))) (term t))

(define (has-select-inout-pair? t)
  (and (equal? (term (select-inout-pair ,t)) (term t))
       (begin (printf "skipping term with inout pair\n")
              #t)))

(test-equal (has-select-inout-pair? (term ((in c1) (out c2) (select (in A) (out B)) (select (out A) (in B) (out A))))) #f)
(test-equal (has-select-inout-pair? (term ((in c1) (out c2) (select (in A) (out B)) (select (out A) (in B) (in A))))) #t)
(test-equal (has-select-inout-pair? (term  ((select (out wH) (in i) (in m) (out V) (in C) (in p) (out I) (in t) (in M) (out M) (out A) (in u) (out Q) (out O) (in n) (out E) (in HI) (out r) (in Z) (in w) (in v) (out u) (out L)) (select (in R) (in EP) (in y) (in g) (in in)) (in z) (select (in O) (in H)) (select (out x) (out d) (out g)))))
            #t)


;; this one ran out of memory.  Is it interesting?

(define difficult-term
  (term ((select (out N)) (select) (select (in E) (out A) (in P) (out w)) (select (out r) (out t) (out P) (in TgD) (in Ne) (in u)) (select (in X) (in c) (out Y) (in q) (out w) (out W) (in i) (out k) (in S) (in A) (out K)) (select (out F)) (select (in F) (in Rg) (in u)) (in b) (in ZTKJ))))

#;(term (compile ,difficult-term))
#;(twiddle-related? difficult-term (term (compile ,difficult-term)) #:max-depth 100)

(random-seed 227264)

#;(redex-check sourcelang prog (or (has-select-inout-pair? (let ([ans (term prog)])
                                                             (printf "~s\n" ans)
                                                             ans))
                                   (twiddle-related? (term prog) (term (compile prog)))) #:attempts 1000)



;; TEST CASE STUFF & JUNK:


#;(let ([p (term ((select (in k) (in K)) (select (in S) (out o) (in P) (in f)) (select (out k) (out o) (out P))))])
    (time (twiddle-related? p (term (compile ,p)) #:max-depth 2)))



(define sample-source-term
  '((select (out C0) (out C2))
    (in C1)
    (select (in C1))
    (select (out C2) (in C2) (in C1))
    (select (out C2) (out C2) (out C1))
    (select (out C1) (in C1))
    (out C0)
    (select (out C0) (out C1) (in C0) (in C0))
    (out C2)
    (select)
    (in C1)
    (out C0)
    (select)
    (select (out C2) (out C0))
    (select (in C2) (out C2) (out C2))
    (out C0)))

#;(test-equal (invariance-holds? sample-source-term (term (compile ,sample-source-term))) #t)
#;(test-equal (invariance-holds? (term ((in C1) (out C1))) (term (compile ((select (in C1) (out C2)) (select (out C1) (in C2)))))) #t)


#|

;; what the heck.  First, let's just try running some programs. randomly generate a source term, find 
;; all its results, compile it, then try that one.


#;(for/list ([i (in-range 13)])
  (generate-term sourcelang prog 30))

;; making it less random to analyze time:

(define (random-source-term size-idx)
  (let ([channels size-idx])
    (for/list ([i (in-range size-idx)]) (random-term-element channels size-idx))))

(define (random-term-element num-channels size-idx)
  (cond [(= (random 2) 1) (random-a num-channels)]
        [else (cons 'select (for/list ([i (in-range (floor (/ size-idx 5)))]) (random-a num-channels)))]))

(define (random-a num-channels)
  (list (if (= (random 2) 1) 'in 'out)
        (string->symbol (format "C~v" (random num-channels)))))

|#


(test-results)