#lang racket

;; Midtgaard & Jensen, ICFP 2009

(require redex racket/pretty
         "util.rkt")

(provide anf lookup fv bind-fv
         μ CaEK Eval
         μc F Fc F-init F-answers
         αx γx Fx Fx-init Fx-answers
         ρ Fρ Fρ-init bind-fv-id
         ≡ ∪· fapp ∪⊗ μ# F#-init F# F#-answers)

;; Figure 2
(define-language anf
  ;; programs
  [(p P) s]
  ;; trivial expressions
  [(t T) c x (fn x s)]
  [c number]
  [(Var x y) variable-not-otherwise-mentioned]
  ;; serious expressions
  [(s s* C) t (let x t s) (t t) (let x (t t) s)]
  ;; Val
  [(w Val) c (close (fn x s) e)]
  ;; Env
  [(e e* Env) • (e x w)]
  ;; K
  [(k k* K) stop (x s e k)]
  ;; machine states (paper calls this 'S' (eg in the definition of 'F' in the beginning of sec 4))
  [M (C Env K)]
  
  ;; set of e
  [E (e ...)]
  
  ;; used for the result of lookup
  [Val/f Val #f]
  
  ;; split machine states
  [S ((C ...) ([C K] ...) (Env ...))]
  
  [(k# K#) stop (x s)]
  [(w# Val#) c (fn x s)]
  [(E# Env#) ((Var (Val# ...)) ...)]
  [(F#) ((s≡ (K# ...)) ...)]
  
  ;; the equivalence class representatives all have this shape (for F#)
  [(s≡) (t t) t])



(define-metafunction anf
  lookup : x Env -> Val/f
  [(lookup x (e x w)) w]
  [(lookup x_1 (e x_2 w)) (lookup x_1 e)]
  [(lookup x •) #f])

(define-metafunction anf
  fv : s -> (x ...)
  [(fv c) ()]
  [(fv x) (x)]
  [(fv (fn x s)) (minus (fv s) x)]
  [(fv (let x t s)) ,(∪ (term (fv t)) (term (minus (fv s) x)))]
  [(fv (t_0 t_1)) ,(∪ (term (fv t_0)) (term (fv t_1)))]
  [(fv (let x (t_0 t_1) s)) ,(∪ (term (fv t_0)) (term (fv t_1)) (term (minus (fv s) x)))])

(define-metafunction anf
  minus : (x ...) x -> (x ...)
  [(minus (x_1 ...) x_2)
   ,(remove* (term (x_2)) (term (x_1 ...)))])

(define (bind-fv t)
  (let ([fv (term (fv ,t))])
    (let loop ([fv fv]
               [i 100])
      (cond
        [(null? fv) t]
        [else `(let ,(car fv) (fn fvx ,i) ,(loop (cdr fv) (+ i 11)))]))))

(define (bind-fv-id t)
  (let ([fv (term (fv ,t))])
    (let loop ([fv fv]
               [i 100])
      (define v (string->symbol (format "fvx~a" i)))
      (cond
        [(null? fv) t]
        [else `(let ,(car fv) (fn ,v ,v) ,(loop (cdr fv) (+ i 1)))]))))


(define-metafunction anf
  μ : T Env -> Val
  [(μ c e) c]
  [(μ x e) 
   Val
   (where Val (lookup x e))]
  [(μ (fn x s) e) (close (fn x s) e)])

(define CaEK
  (reduction-relation 
   anf #:domain M
   [--> (t e (x s* e* k*))
        (s* (e* x (μ t e)) k*)
        "rule1"]
   [--> ((let x t s) e k)
        (s (e x (μ t e)) k)
        "rule2"]
   [--> ((t_0 t_1) e k)
        (s* (e* x w) k)
        (where (close (fn x s*) e*) (μ t_0 e))
        (where w (μ t_1 e))
        "rule3"]
   [--> ((let x (t_0 t_1) s) e k)
        (s* (e* y w) (x s e k))
        (where (close (fn y s*) e*) (μ t_0 e))
        (where w (μ t_1 e))
        "rule4"]))

;; we map cn1 to 1, etc
(define example
  (term (let f (fn x x) (let a1 (f 1) (let a2 (f 2) a2)))))

(define (start p)
  (term (,p • (xr xr • stop))))

(define extract-answer
  (term-match/single anf
                     [(xr (• xr w) stop)
                      (term w)]))

(define (Eval p)
  (define final-states (apply-reduction-relation* CaEK (start p)))
  (unless (= 1 (length final-states))
    (error 'Eval "~s reduced to multiple final states: ~s" p final-states))
  (extract-answer (car final-states)))
  
;; Section 4

;; Figure 3

(define-metafunction anf
  μc : T (Env ...) -> (Val ...)
  [(μc c E) (c)]
  [(μc x (e ...)) 
   ,(filter values
            (for/list ([env (term (e ...))])
              (term (lookup x ,env))))]
  [(μc (fn x s) (e ...)) ((close (fn x s) e) ...)])


;; beginning of section 4, inset text
(define ((F p) S)
  (∪ (F-init p)
     (apply append (map (λ (s) (apply-reduction-relation CaEK s)) S))))

;; Figure 3b
(define ((Fc p) S)
  (∪ (F-init p)
     (set-of anf 
             (s* (e* x w) k*)
             (t e (x s* e* k*)) ∈ ,S
             w ∈ (μc t (e)))
     
     (set-of anf 
             (s (e x w) k)
             ((let x t s) e k) ∈ ,S
             w ∈ (μc t (e)))
     
     (set-of anf 
             (s* (e* x w) k)
             ((t_0 t_1) e k) ∈ ,S
             (close (fn x s*) e*) ∈ (μc t_0 (e))
             w ∈ (μc t_1 (e)))
     
     (set-of anf 
             (s* (e* y w) (x s e k))
             ((let x (t_0 t_1) s) e k) ∈ ,S
             (close (fn y s*) e*) ∈ (μc t_0 (e))
             w ∈ (μc t_1 (e)))))

(define (F-init p)
  (term ((,p • (xr xr • stop)))))

;; F-answers : (listof (C Env K)) -> (listof w)
;; extracts the value of final configurations from a set for the collecting semantics
(define F-answers
  (let ([get-ans
         (term-match/single anf 
                            [(xr (• xr w) stop) (term w)]
                            [any #f])])
    (λ (set)
      (filter values (map get-ans set)))))

(define-metafunction anf
  αx : (M ...) -> ((C ...) ([C K] ...) (Env ...))
  [(αx (M ...))
   (,(clean-set (map first (term (M ...))))
    ,(clean-set (map (λ (m) (list (first m) (third m))) (term (M ...))))
    ,(clean-set (map second (term (M ...)))))])

(define-metafunction anf
  γx : ((C ...) ([C K] ...) (Env ...)) -> (M ...)
  [(γx (any_c any_f any_e))
   ,(for*/list ([s (term any_c)]
                [sk (term any_f)]
                [e (term any_e)]
                #:when (equal? s (first sk)))
      (list s e (second sk)))])

(define (Fx-init p)
  (term ((,p) ((,p (xr xr • stop))) (•))))

(define ((Fx p) C-F-E)
  (apply ∪x
         (append
          (list (Fx-init p))
          (set-of anf
                  ((s*) ((s* k*)) ((e* x w)))
                  (t (t (x s* e* k*)) e) ∈x ,C-F-E
                  w ∈ (μc t (e)))
          (set-of anf
                  ((s) ((s k)) ((e x w)))
                  ((let x t s) ((let x t s) k) e) ∈x ,C-F-E
                  w ∈ (μc t (e)))
          (set-of anf
                  ((s*) ((s* k)) ((e* x w)))
                  ((t_0 t_1) ((t_0 t_1) k) e) ∈x ,C-F-E
                  (close (fn x s*) e*) ∈ (μc t_0 (e))
                  w ∈ (μc t_1 (e)))
          (set-of anf
                  ((s*) ((s* (x s e k))) ((e* y w)))
                  ((let x (t_0 t_1) s) ((let x (t_0 t_1) s) k) e) ∈x ,C-F-E
                  (close (fn y s*) e*) ∈ (μc t_0 (e))
                  w ∈ (μc t_1 (e))))))

(define (Fx-answers set)
  (F-answers (term (γx ,set))))

(define (ρ C-F-E)
  (define s-ks '())
  (define es '())
  
  (define (add-e e)
    (set! es (cons e es))
    (match e
      [`(,e1 ,x (close ,fn ,e2))
       (add-e e1)
       (add-e e2)]
      [`(,e ,x ,w)
       (add-e e)]
      [`•
       (void)]))
  
  (match C-F-E
    [`(,C ,F ,E)
     ;; build the s-ks
     (for ([s-k (in-list F)])
       (let loop ([s-k s-k])
         (set! s-ks (cons s-k s-ks))
         (match s-k
           [`(,s (,x ,s* ,e ,k))
            (loop `(,s* ,k))]
           [`(,s stop)
            (void)])))

     ;; find the e's in the F field
     (for ([s-k (in-list F)])
       (match s-k
         [`(,s ,k)
          (let loop ([k k])
            (match k
              [`(,x ,s ,e ,k)
               (add-e e)
               (loop k)]
              [`stop
               (void)]))]))
     
     ;; find the e's in the E
     (for ([e (in-list E)])
       (add-e e))

     `(,C ,(clean-set s-ks) ,(clean-set es))]))
  
(define (Fρ-init p)
  ;; should be the same as (ρ (Fx-init p)), but
  ;; the paper doesn't write it that way
  `((,p) ((,p (xr xr • stop)) (xr stop)) (•)))

(define ((Fρ p) C-F-E)
  (apply ∪x
         (append
          (list (Fρ-init p))
          (set-of anf
                  ,(ρ (term ((s*) ((s* k*)) ((e* x w)))))
                  (t (t (x s* e* k*)) e) ∈x ,C-F-E
                  w ∈ (μc t (e)))
          (set-of anf
                  ,(ρ (term ((s) ((s k)) ((e x w)))))
                  ((let x t s) ((let x t s) k) e) ∈x ,C-F-E
                  w ∈ (μc t (e)))
          (set-of anf
                  ,(ρ (term ((s*) ((s* k)) ((e* x w)))))
                  ((t_0 t_1) ((t_0 t_1) k) e) ∈x ,C-F-E
                  (close (fn x s*) e*) ∈ (μc t_0 (e))
                  w ∈ (μc t_1 (e)))
          (set-of anf
                  ,(ρ (term ((s*) ((s* (x s e k))) ((e* y w)))))
                  ((let x (t_0 t_1) s) ((let x (t_0 t_1) s) k) e) ∈x ,C-F-E
                  (close (fn y s*) e*) ∈ (μc t_0 (e))
                  w ∈ (μc t_1 (e))))))

(define-metafunction anf
  ≡ : s -> s
  [(≡ (let x t s)) (≡ s)]
  [(≡ (let x (t_0 t_1) s)) (≡ s)]
  [(≡ (t_0 t_1)) (t_0 t_1)]
  [(≡ t) t])

;; functions that map to sets are represented as
;; assoc lists for their supprt, 
;; using 'clean-set' on the range lists
;; eg if
;;   f(x) = {1, 2, 3}
;;   f(y) = {4}
;;   f(_) = {}   // _ is "anything else"
;; then f is represented as:
;;   ((x (1 2 3)) (y (4)))
(define-metafunction anf
  [(∪· any ...)
   ,(let ([ht (make-hash)])
      (for ([arg (in-list (term (any ...)))])
        (for ([frame (in-list arg)]
              [arg-i (in-naturals)])
          (unless (and (list? frame)
                       (= 2 (length frame)))
            (error '∪· "arg ~a is ~s but should be a list of length 2, all args: ~s"
                   arg-i frame (term (any ...))))
          (define dom (list-ref frame 0))
          (define rng (list-ref frame 1))
          (unless (null? rng)
            (hash-set! ht dom (append rng (hash-ref ht dom '()))))))
      (sort (hash-map ht (λ (k v) (list k (clean-set v))))
            string<=?
            #:key (λ (x) (format "~s" (car x)))))])

(define-metafunction anf
  [(fapp (any_1 ... (any_2 (any_3 ...)) any_4 ...) any_2)
   (any_3 ...)]
  [(fapp any_1 any_2)
   ()])

(define (∪⊗ arg1 . args)
  (define all-args (cons arg1 args))
  (for ([x (in-list all-args)]
        [y (in-naturals)])
    (unless (and (list? x)
                 (= (length x) 3))
      (error '∪⊗ 
             "expected all arguments to be lists of length 3, but argument ~a is ~s"
             y
             x)))
  (define fst (map first all-args))
  (define snd (map second all-args))
  (define trd (map third all-args))
  (list (apply ∪ fst)
        (term (∪· ,@snd))
        (term (∪· ,@trd))))

(define (F#-init p)
  (term ((,p) 
         (((≡ ,p) ((xr xr)))
          ((≡ xr) (stop)))
         ())))

(define-metafunction anf
  μ# : t E# -> (w# ...)
  [(μ# c E#) (c)]
  [(μ# x E#) (fapp E# x)]
  [(μ# (fn x s) E#) ((fn x s))])

(define C? (redex-match anf (s ...)))
(define F#? (redex-match anf F#))
(define E#? (redex-match anf E#))

(define (chk-C-F#-E# what C-F#-E#)
  (match C-F#-E#
    [`(,C ,F#, E#)
     (unless (C? C) (error 'F# "~a has illformed C: ~s" what C))
     (unless (F#? F#) (error 'F# "~a has illformed F#: ~s" what F#))
     (unless (E#? E#) (error 'F# "~a has illformed E#: ~s" what E#))
     C-F#-E#]))

(define ((F# p) C-F#-E#)
  (match (chk-C-F#-E# 'input C-F#-E#)
    [`(,C ,F#, E#)
     (chk-C-F#-E# 
      'result
      (apply ∪⊗
             (append
              (list (F#-init p))
              (set-of anf
                      ((s*) ,F# (∪· ,E# ((x (μ# t ,E#)))))
                      t ∈ ,C
                      (x s*) ∈ (fapp ,F# (≡ t)))
              (set-of anf
                      ((s) ,F# (∪· ,E# ((x (μ# t ,E#)))))
                      (let x t s) ∈ ,C)
              (set-of anf
                      ((s*) (∪· ,F# (((≡ s*) (fapp ,F# (≡ (t_0 t_1)))))) (∪· ,E# ((x (μ# t_1 ,E#)))))
                      (t_0 t_1) ∈ ,C
                      (fn x s*) ∈ (μ# t_0 ,E#))
              (set-of anf
                      ((s*) (∪· ,F# (((≡ s*) ((x s))))) (∪· ,E# ((y (μ# t_1 ,E#)))))
                      (let x (t_0 t_1) s) ∈ ,C
                      (fn y s*) ∈ (μ# t_0 ,E#)))))]))

(define stop-kont-frame? (redex-match anf (s≡ (k#_1 ... stop k#_2 ...))))

(define (F#-answers C-F#-E#)
  (match C-F#-E#
    [`(,C ,F#, E#)
     (define final-s≡s (map car (filter stop-kont-frame? F#)))
     (apply append (map (λ (var)
                          (if (symbol? var)
                              (term (fapp ,E# ,var))
                              (list var)))
                        final-s≡s))])) 

